/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include "etherxd.h"

/* this is the intake feed from transports to deliver epackets */
void etherxd_send(epacket ep)
{
    ehandler eh;

    log_debug(ZONE,"etherxd_send()");

    if(ep == NULL)
        return;

    eh = ehandler_list_get(etherx__handlers,ep->to,ep->ns);
    if(eh != NULL)
    {
        etherxd_deliver(eh,ep);
        return;
    }

    if(ep->type == EPACKET_ERROR)
    {
        log_error("etherxd","can't bounce a packet to %s from %s, dropping",ep->to, ep->from);
        xmlnode_free(ep->x);
        return;
    }

    eh = ehandler_list_get(etherxd__connections,ep->to,ep->ns);
    if(eh == NULL)
        eh = etherxd_netout_connect(ep->to, ep->ns);

    log_debug(ZONE,"deliver to netout %s",eh->host);

    eh->count++;
    etherxd_netout_write(eh->t, ep); /* our special handler */
}

/* sends epacket to the trasports ehandler, encoding where needed for socket delivery */
void etherxd_deliver(ehandler eh, epacket ep)
{
    eh->count++;

    if(eh->h != NULL)
    {
        log_debug(ZONE,"deliver to transport(INPROC) %s",eh->host);
        (eh->h)(ep);
    }else{
        log_debug(ZONE,"deliver to transport(SOCKET) %s",eh->host);
        xmlstream_write(eh->t, epacket_serialize(ep));
    }
}

/* this is the callback on new xmlstreams, it checks/sets-up the stream
 * and determines if this is a normal connect or a new socket-based transport
 */
void etherxd_negotiate(thread t, xmlnode in, int type)
{
    char *ns, *to;
    tstream ts = (tstream)(t->data);
    xmlnode x;
    ehandler eh, ehnew;

    if(type != XMLSTREAM_ROOT)
        return;

    log_debug(ZONE,"manager_negotiate socket %d root %s",ts->fd,xmlnode2str(in));

    ns = xmlnode_get_attrib(in, "xmlns");
    to = xmlnode_get_attrib(in, "to");

    if(type == XMLSTREAM_ERROR)
    {
        log_debug(ZONE,"negotiated: stream error");
        xmlstream_write_header(t, xmlstream_header(NULL, NULL, to));
        xmlstream_error(t, "You didn't send valid XML.");
        return;
    }

    if(ns == NULL)
    {
        log_debug(ZONE,"negotiated: stream error, no namespace");
        xmlstream_write_header(t, xmlstream_header(NULL, NULL, to));
        xmlstream_error(t, "You didn't specify a default xmlns='' namespace.");
        return;
    }

    if(xmlnode_get_attrib(in, "xmlns:etherx") != NULL && xmlnode_get_attrib(in,"etherx:secret") != NULL)
    { /* it's trying to be a transport */
        spool s;

        log_debug(ZONE,"negoitiated: new transport");

        s = spool_new(xmlnode_pool(in));
        spooler(s, xmlnode_get_attrib(in, "id"), etherx__secret, s);
        if(j_strcmp(xmlnode_get_attrib(in,"etherx:secret"),shahash(spool_print(s))) != 0)
        {
            xmlstream_write_header(t, xmlstream_header(NULL, NULL, to));
            xmlstream_error(t, "Invalid Secret, transport access denied");
            return;
        }

        /* find any matching transport */
        eh = ehandler_list_get(etherx__handlers,xmlnode_get_attrib(in,"from"),ns);

        /* you can't replace an in-process transport! */
        if(eh != NULL && eh->h != NULL)
        {
            xmlstream_write_header(t, xmlstream_header(NULL, NULL, to));
            xmlstream_error(t, "Transport access denied");
            return;
        }

        x = xmlstream_header(ns, NULL, to);
        xmlnode_put_attrib(x, "xmlns:etherx",ETHERX_NAMESPACE);
        xmlstream_write_header(t, x);

        /* it's a transport, let them manage it */
        ts->data = (void *)etherxd_transport_onNode;
        pool_cleanup(t->p, etherxd_transport_cleanup, (void *)t);

        if(eh == NULL)
        {
            /* create it's ehandler as a transport or sender only */
            if(xmlnode_get_attrib(in,"from") != NULL)
            {
                eh = ehandler_new(NULL, xmlnode_get_attrib(in,"from"), ns);
                etherx__handlers = ehandler_list_add(etherx__handlers,eh);
            }else{
                eh = ehandler_new(NULL, ts->host, ns);
            }
            eh->children = NULL;
        }else{
            xmlstream_error(eh->t, "Replaced by new transport connection");
        }
        ts->data_aux1 = (void *)eh;
        eh->t = t;

    }else{
        /* it's connecting to a transport, find and hook together */
        if(to == NULL)
        {
            xmlstream_write_header(t, xmlstream_header(NULL, NULL, NULL));
            xmlstream_error(t, "You didn't send your to='host' attribute.");
            return;
        }

        log_debug(ZONE,"negoitiated: incoming connection");
        eh = ehandler_list_get(etherx__handlers,to,ns);
        if(eh == NULL)
        {
            xmlstream_write_header(t, xmlstream_header(NULL, NULL, to)); /* important: if their namespace is invalid, don't claim to be talking in it */
            xmlstream_error(t, "Unable to locate server/namespace.");
        }else{
            xmlstream_write_header(t, xmlstream_header(ns, NULL, to));
            /* setup netin_* to take care of it */
            ts->data_aux5 = (void *)eh;
            ts->data = (void *)etherxd_netin_onNode;
            pool_cleanup(t->p, etherxd_netin_cleanup, (void *)t);

            /* save an entry on this transports list of connections */
            ehnew = ehandler_new(NULL, ts->host, ns);
            ehnew->t = t;
            eh->children = ehandler_list_add(eh->children,ehnew);
            ts->data_aux1 = (void *)ehnew;
        }
    }
}

/* this is the callback on the listener thread whenever there is a new connect */
void etherxd_onConnect(thread t, int sock, struct sockaddr_in sa)
{
    thread tnew;

    log_debug(ZONE,"connection established (fd: %d, ip: %s, port: %d)",
              sock, inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));
    tnew = xmlstream_new(sock, inet_ntoa(sa.sin_addr), (xmlstream_onNode)etherxd_negotiate);
}

int etherxd_start(void)
{
    log_debug(ZONE,"local etherxd attempting to start");

    /* already started */
    if(etherx__daemon != NULL)
        return 1;

    etherx__daemon = tlisten_new(5269,NULL,etherxd_onConnect,NULL);

    if(etherx__daemon != NULL)
        return 1;
    else
        return 0;
}

