/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include <etherx.h>
#include "config.h"
#ifdef HAVE_LIBADNS
#include <adns.h>
#endif
#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif

thread tstream_connect(int port, char *host, tstream_handler_read read, tstream_handler_err err, void *data)
{
    return _tstream_new(0, port, host, read, err, data);
}

thread tstream_new(int fd, char *host, tstream_handler_read read, tstream_handler_err err, void *data)
{
    return _tstream_new(fd, 0, host, read, err, data);
}

void tstream_rw(tstream ts, void *reader, void *writer)
{
    if(reader != NULL)
        ts->reader = reader;
    if(writer != NULL)
        ts->writer = writer;
}

int _tstream_pth_read(tstream ts, void *buf, size_t count)
{
    return pth_read(ts->fd, buf, count);
}

int _tstream_pth_write(tstream ts, const void *buf, size_t count)
{
    return pth_write(ts->fd, buf, count);
}

thread _tstream_new(int fd, int port, char *host, tstream_handler_read read, tstream_handler_err err, void *data)
{
    thread t;
    tstream ts;
    pth_attr_t attr;

    log_debug(ZONE,"tstream_new");

    t = thread_new();
    ts = pmalloc(t->p, sizeof(_tstream));
    memset(ts, '\0',sizeof(_tstream));

    ts->write_mp = pth_msgport_create("tstream");
    ts->h_read = read;
    ts->h_err = err;
    ts->data = data;
    ts->host = pstrdup(t->p,host);
    ts->fd = fd;
    ts->port = port;

    tstream_rw(ts,_tstream_pth_read,_tstream_pth_write);
    t->data = (void *)ts;

    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(attr, PTH_ATTR_STACK_SIZE, etherx_stack_default);
    t->id = pth_spawn(attr, _tstream_main, (void *)t);
    pth_attr_destroy(attr);
    /* pth_yield(NULL); let pth start the thread */

    return t;
}

void tstream_write(thread t, void *buffer, int bufsz)
{
    tstream_buffer b;
    pool p;

    if(buffer == NULL || bufsz == 0 || t == NULL)
        return;

    if(bufsz < 0) /* send a -1 size and it will figure it out for you */
        bufsz = strlen(buffer);

    p = pool_heap(bufsz + 32);
    b = pmalloc(p,sizeof(struct tstream_buffer_struct));

    b->data = pstrdup(p,buffer);
    b->data_sz = bufsz;
    b->sent = 0;
    b->obj = b->next = NULL;
    b->p = p;

    tstream_write_buffer(t, b);
}

void tstream_close(thread t)
{
    tstream ts = (tstream)(t->data);

    log_debug(ZONE,"asking socket to close");
    ts->close = 1;
    if(ts->write == NULL)
    {
        log_debug(ZONE,"closing socket");
        if(t->id == pth_self())
            pth_exit(NULL);
        else
            pth_cancel(t->id);
    }
}

void tstream_write_buffer(thread t, tstream_buffer buffer)
{
    tstream ts = (tstream)(t->data);

    log_debug(ZONE,"tstream_write_buffer");

    if(ts->write == NULL)
        ts->write = buffer;
    else
        pth_msgport_put(ts->write_mp, (pth_message_t *)buffer);

    if(ts->fd > 0 && t->id != pth_self())
        _tstream_write(t); /* write the data pronto if we can from this thread... if same thread, we'll brb */
    if(ts->write != NULL)
        _tstream_cpu(t); /* if there's data left for writing, the thread has to get cpu time to handle it */
}

/* give the tstream some cpu time, typically to flag that there is data waiting to be written */
void _tstream_cpu(thread t)
{
    tstream ts = (tstream)(t->data);
    tstream_buffer b;
    pool p;

    p = pool_new();
    b = pmalloc(p,sizeof(struct tstream_buffer_struct));

    b->data = NULL;
    b->data_sz = 0;
    b->sent = 0;
    b->obj = NULL;
    b->p = p;

    pth_msgport_put(ts->write_mp, (pth_message_t *)b);
}

void tstream_oob_buffer(thread t, tstream_buffer buffer)
{
    tstream ts = (tstream)(t->data);
    tstream_buffer curr = ts->oob;

    log_debug(ZONE,"tstream_oob_buffer");

    buffer->next = NULL;
    if(curr == NULL)
    {
        ts->oob = buffer;
        return;
    }

    while(curr->next != NULL)
        curr = curr->next;
    curr->next = buffer;
}

void tstream_oob_flush(thread t)
{
    tstream ts = (tstream)(t->data);

    log_debug(ZONE,"tstream_oob_flush");

    while(ts->oob != NULL)
    {
        pth_msgport_put(ts->write_mp, (pth_message_t *)ts->oob);
        ts->oob = ts->oob->next;
    }

    if(ts->write == NULL)
    {
        ts->write = (tstream_buffer)(pth_msgport_get(ts->write_mp));
        if(ts->write != NULL)
            _tstream_cpu(t);
    }
}

void *_tstream_main(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    pth_event_t mpevt, readevt, writeevt, evtring;
    int data_sz;
    char data[1024];

    log_debug(ZONE,"tstream_main: startup");

    pth_cleanup_push(_tstream_cleanup, (void *)t);

    if(ts->h_read != NULL)
        (ts->h_read)(t, NULL, 0); /* signal an INIT to the read event */

    if(ts->fd == 0) /* try to connect */
        ts->fd = _tstream_connect(ts->port, ts->host);

    if(ts->fd <= 0) /* uhoh, bad fd! */
        pth_exit(NULL);

    if(ts->h_read == NULL)
    {
        readevt = NULL; /* no event to handle reads, hence no reads */
    }else{
        readevt = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE, ts->fd);
        pool_cleanup(t->p, _tstream_cleanup_evt, (void *)readevt);
    }
    writeevt = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE, ts->fd);
    pth_cleanup_push(_tstream_cleanup_evt, (void *)writeevt);
    mpevt = pth_event(PTH_EVENT_MSG,ts->write_mp);
    pth_cleanup_push(_tstream_cleanup_evt, (void *)mpevt);

    evtring = pth_event_concat(mpevt, readevt, NULL);
    while(pth_wait(evtring) > 0)
    {
        if(pth_event_occurred(readevt))
        { /* discussion: better to loop fetching all data waiting on socket
               * or return control to pth and let it come back?
               * If we return control, other waiting threads get a chance to breathe
               * if we read as much as possible, faster throughput for fast connections?
               * 1/19/00 jer: I decided to change this and let pth have control after every read, DoS prevention
               */
            log_debug(ZONE,"tstream read event");
            /*            do
                        {*/
            data_sz = (ts->reader)(ts, data, 1023);
            if(data_sz > 0)
            {
                data[data_sz] = '\0';
                (ts->h_read)(t, data, data_sz);
            }else{
                log_warn("etherx","error '%s' reading from socket '%s'",strerror(errno),ts->host);
                pth_exit(NULL);
            }
            /*            } while(data_sz == 1023); */
        }
        if(pth_event_occurred(mpevt) || pth_event_occurred(writeevt))
        {
            log_debug(ZONE,"tstream write event %d",ts->close);
            _tstream_write(t);

            if(ts->write == NULL)
            {
                pth_event_isolate(writeevt);
                evtring = pth_event_concat(mpevt, readevt, NULL);
            }else{
                evtring = pth_event_concat(mpevt, writeevt, readevt, NULL);
            }
        }
    }
    return NULL;
}

void _tstream_write(thread t)
{
    tstream ts = (tstream)(t->data);
    int data_sz;

    while(ts->write != NULL)
    {
        log_debug(ZONE,"tstream_write %d blocks",ts->write->data_sz);
        data_sz = (ts->writer)(ts, ts->write->data + ts->write->sent, ts->write->data_sz - ts->write->sent);
        if(data_sz == -1 && errno == EAGAIN)
            data_sz = 0; /* so we retry if the socket is constipated */
        if(data_sz < 0)
        {
            /* print error if we weren't expecting to close */
            if(ts->close == 0)
                log_warn("etherx","error '%s' writing to socket '%s'",strerror(errno), ts->host);
            /* we're outta here :) */
            ts->close = 1;
            break;
        }
        if((data_sz + ts->write->sent) == ts->write->data_sz)
        {
            /* this packets empty, crush and move on */
            pool_free(ts->write->p);
            ts->write = (tstream_buffer)(pth_msgport_get(ts->write_mp));
        }else{
            /* not all of the data was sent, socket must be full, come back later */
            ts->write->sent += data_sz;
            return;
        }
    }

    if(ts->close) /* die! die! die! */
    {
        log_debug(ZONE,"closing socket");
        if(t->id == pth_self())
            pth_exit(NULL);
        else
            pth_cancel(t->id);
    }
}

#ifdef HAVE_LIBADNS
int _tstream_connect(int port, char *host)
{
    int maxfd, fd, rrn;
    void *rrp;
    adns_state ads;
    adns_query aq;
    adns_answer *ans;
    fd_set rfds, wfds, efds;
    struct timeval *tv, tvbuf;
    struct in_addr *iap, ia;
    adns_rr_inthostaddr *rriha;
    adns_rr_hostaddr *rrha;
    struct sockaddr_in sa;

    log_debug(ZONE,"tstream_connect: ADNS[%s:%d]",host,port);

    if((ia.s_addr = inet_addr(host)) != INADDR_NONE)
        return _tstream_socket(port, &ia);

    adns_init(&ads, adns_if_noerrprint|adns_if_noautosys|adns_if_nosigpipe|adns_if_noenv, 0);

    /* check for A first */
    if(adns_submit(ads, host, adns_r_a, 0, NULL, &aq))
        return -1;
    while(1)
    {
        FD_ZERO(&rfds);
        FD_ZERO(&wfds);
        FD_ZERO(&efds);
        maxfd = 1;
        tv = NULL;

        adns_beforeselect(ads, &maxfd, &rfds, &wfds, &efds, &tv, &tvbuf,0);
        pth_select(maxfd, &rfds, &wfds, &efds, tv);
        adns_afterselect(ads, maxfd, &rfds,&wfds,&efds, 0);

        log_debug(ZONE,"tstream_connect: ADNS A lookup loop");

        if(!adns_check(ads, &aq, &ans, NULL))
            break;
    }

    if(ans->status == adns_s_ok && ans->nrrs > 0)
    {
        log_debug(ZONE,"tstream_connect: ADNS found some responses, attempting to connect");
        fd = 0;
        for(rrn = ans->nrrs, rrp = ans->rrs.untyped + (ans->rrsz*(ans->nrrs-1)); rrn > 0 && fd <= 0;  rrn--, rrp -=ans->rrsz)
        {
            iap = (struct in_addr *)rrp;
            fd = _tstream_socket(port, iap);
        }
        if(fd > 0)
        {
            log_debug(ZONE,"tstream_connectd: ADNS[%s]",inet_ntoa(*iap));
            adns_finish(ads);
            return fd;
        }
    }

    /* check MX next */
    if(adns_submit(ads, host, adns_r_mx, 0, NULL, &aq))
        return -1;
    while(1)
    {
        FD_ZERO(&rfds);
        FD_ZERO(&wfds);
        FD_ZERO(&efds);
        maxfd = 1;
        tv = NULL;

        adns_beforeselect(ads, &maxfd, &rfds, &wfds, &efds, &tv, &tvbuf,0);
        pth_select(maxfd, &rfds, &wfds, &efds, tv);
        adns_afterselect(ads, maxfd, &rfds,&wfds,&efds, 0);

        log_debug(ZONE,"tstream_connect: ADNS MX lookup loop");

        if(!adns_check(ads, &aq, &ans, NULL))
            break;
    }

    if(ans->status == adns_s_ok && ans->nrrs > 0)
    {
        log_debug(ZONE,"tstream_connect: ADNS found some responses, attempting to connect");
        fd = 0;
        for(rrn = ans->nrrs, rrp = ans->rrs.untyped + (ans->rrsz*(ans->nrrs-1)); rrn > 0 && fd <= 0;  rrn--, rrp -=ans->rrsz)
        {
            rriha = (adns_rr_inthostaddr *)rrp;
            log_debug(ZONE,"checking mx %u ",rriha->i);
            rrha = &rriha->ha;
            if(rrha->naddrs <= 0)
            {
                if((ia.s_addr = inet_addr(rrha->host)) != INADDR_NONE)
                    iap = &ia;
                else
                    continue;
            }
            else
            {
                sa = rrha->addrs->addr.inet;
                iap = &sa.sin_addr;
            }

            fd = _tstream_socket(port, iap);
        }
        if(fd > 0)
        {
            log_debug(ZONE,"tstream_connectd: ADNS[%s]",inet_ntoa(*iap));
            adns_finish(ads);
            return fd;
        }
    }

    adns_finish(ads);
    return -1;
}
#else
int _tstream_connect(int port, char *host)
{
    struct hostent *hp;
    struct in_addr ia;

    log_debug(ZONE,"tstream_connect: HOST[%s]",host);

    if((ia.s_addr = inet_addr(host)) != INADDR_NONE)
        return _tstream_socket(port, &ia);

    hp = gethostbyname(host);
    if(hp == NULL)
        return -1;
    return _tstream_socket(port, (struct in_addr *)*hp->h_addr_list);
}
#endif /* HAVE_LIBADNS */

void _tstream_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);

    log_debug(ZONE,"tstream_cleanup");

    if(ts->oob != NULL)
        tstream_oob_flush(t);

    while(ts->write != NULL)
    {
        if(ts->h_err != NULL)
            (ts->h_err)(t, ts->write);
        pool_free(ts->write->p);
        ts->write = (tstream_buffer)(pth_msgport_get(ts->write_mp));
    }

    pth_msgport_destroy(ts->write_mp);
    close(ts->fd);
    thread_free(t); /* free's the pool and everything else */
}

void _tstream_cleanup_evt(void *arg)
{
    pth_event_free((pth_event_t)arg,PTH_FREE_THIS);
}

int _tstream_socket(int port, struct in_addr *iap)
{
    struct sockaddr_in sa;
    int s, flag = 1;
    pth_event_t evt;

    log_debug(ZONE,"tstream_socket %d %s",port,inet_ntoa(*iap));

    bzero((void *)&sa,sizeof(struct sockaddr_in));

    if((s = socket(AF_INET,SOCK_STREAM,0)) < 0)
        return(-1);
    if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*)&flag, sizeof(flag)) < 0)
        return(-1);

    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);
    sa.sin_addr.s_addr = iap->s_addr;
    evt = pth_event(PTH_EVENT_TIME, pth_timeout(10,0));
    pth_fdmode(s,PTH_FDMODE_NONBLOCK);
    if(pth_connect_ev(s,(struct sockaddr*)&sa,sizeof sa, evt) < 0)
    {
        close(s);
        return(-1);
    }

    return s;
}

